/*===========================================================================
  This library is released under the MIT license. See RefPtr.html
  for further information and documentation.

Copyright (c) 2008 Juha Nieminen

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
=============================================================================*/

#ifndef INCLUDE_REF_PTR_HH
#define INCLUDE_REF_PTR_HH

#include <memory>

template<typename Obj_t, typename Allocator = std::allocator<Obj_t> >
class RefPtr
{
 public:
    RefPtr();

    enum Inst { instantiate };
    RefPtr(Inst, const Allocator& = Allocator());

    // All these constructors should be changed to a single variadic
    // template when the next C++ standard is ratified.
    template<typename T1>
    RefPtr(const T1&, const Allocator& = Allocator());

    template<typename T1, typename T2>
    RefPtr(const T1&, const T2&, const Allocator& = Allocator());

    template<typename T1, typename T2, typename T3>
    RefPtr(const T1&, const T2&, const T3&, const Allocator& = Allocator());

    template<typename T1, typename T2, typename T3, typename T4>
    RefPtr(const T1&, const T2&, const T3&, const T4&,
           const Allocator& = Allocator());

    template<typename T1, typename T2, typename T3, typename T4, typename T5>
    RefPtr(const T1&, const T2&, const T3&, const T4&, const T5&,
           const Allocator& = Allocator());

    template<typename T1, typename T2, typename T3, typename T4, typename T5,
             typename T6>
    RefPtr(const T1&, const T2&, const T3&, const T4&, const T5&, const T6&,
           const Allocator& = Allocator());

    ~RefPtr();
    RefPtr(const RefPtr&);
    RefPtr& operator=(const RefPtr&);

    Obj_t* operator->();
    const Obj_t* operator->() const;

    Obj_t& operator*();
    const Obj_t& operator*() const;

    bool operator==(const RefPtr&) const;
    bool operator!=(const RefPtr&) const;

    Obj_t* get() const;
    operator bool() const;
    bool isShared() const;



//--------------------------------------------------------------------------
 private:
    struct RefCount: public Allocator
    {
        size_t refCount;

        RefCount(const Allocator& a): Allocator(a), refCount(1) {}
    };

    template<typename Obj_t2>
    struct Data: public RefCount
    {
        Obj_t2 data;

        Data(const Allocator& a): RefCount(a) {}

        // These should also be replaced with a single variadic template...
        template<typename T1>
        Data(const T1& v1, const Allocator& a):
            RefCount(a), data(v1) {}

        template<typename T1, typename T2>
        Data(const T1& v1, const T2& v2, const Allocator& a):
            RefCount(a), data(v1, v2) {}

        template<typename T1, typename T2, typename T3>
        Data(const T1& v1, const T2& v2, const T3& v3, const Allocator& a):
            RefCount(a), data(v1, v2, v3) {}

        template<typename T1, typename T2, typename T3, typename T4>
        Data(const T1& v1, const T2& v2, const T3& v3, const T4& v4,
             const Allocator& a):
            RefCount(a), data(v1, v2, v3, v4) {}

        template<typename T1, typename T2, typename T3, typename T4,
                 typename T5>
        Data(const T1& v1, const T2& v2, const T3& v3, const T4& v4,
             const T5& v5, const Allocator& a):
            RefCount(a), data(v1, v2, v3, v4, v5) {}

        template<typename T1, typename T2, typename T3, typename T4,
                 typename T5, typename T6>
        Data(const T1& v1, const T2& v2, const T3& v3, const T4& v4,
             const T5& v5, const T6& v6, const Allocator& a):
            RefCount(a), data(v1, v2, v3, v4, v5, v6) {}
    };

    RefCount* dataPtr;
    static void(*deleterFunc)(RefCount*);

    static void deallocateData(RefCount* data)
    {
        typedef Data<Obj_t> Data_t;
        typedef typename Allocator::template rebind<Data_t>::other
            DataAllocator;

        Data_t* ptr = static_cast<Data_t*>(data);
        ptr->Data_t::~Data_t();
        DataAllocator(*ptr).deallocate(ptr, 1);
    }

    void allocateData(const Allocator& alloc)
    {
        typedef Data<Obj_t> Data_t;
        typedef typename Allocator::template rebind<Data_t>::other
            DataAllocator;

        dataPtr = DataAllocator(alloc).allocate(1);
        deleterFunc = &deallocateData;
    }

    inline void decRefCount()
    {
        if(dataPtr && --(dataPtr->refCount) == 0)
            deleterFunc(dataPtr);
    }
};

template<typename Obj_t, typename Allocator>
void(*RefPtr<Obj_t, Allocator>::deleterFunc)(RefCount*) = 0;

template<typename Obj_t, typename Allocator>
RefPtr<Obj_t, Allocator>::RefPtr():
    dataPtr(0)
{}

template<typename Obj_t, typename Allocator>
RefPtr<Obj_t, Allocator>::RefPtr(Inst, const Allocator& alloc)
{
    allocateData(alloc);
    new(static_cast<Data<Obj_t>*>(dataPtr)) Data<Obj_t>(alloc);
}

template<typename Obj_t, typename Allocator>
template<typename T1>
RefPtr<Obj_t, Allocator>::RefPtr(const T1& v1, const Allocator& alloc)
{
    allocateData(alloc);
    new(static_cast<Data<Obj_t>*>(dataPtr)) Data<Obj_t>(v1, alloc);
}

template<typename Obj_t, typename Allocator>
template<typename T1, typename T2>
RefPtr<Obj_t, Allocator>::RefPtr(const T1& v1, const T2& v2,
                                  const Allocator& alloc)
{
    allocateData(alloc);
    new(static_cast<Data<Obj_t>*>(dataPtr)) Data<Obj_t>(v1, v2, alloc);
}

template<typename Obj_t, typename Allocator>
template<typename T1, typename T2, typename T3>
RefPtr<Obj_t, Allocator>::RefPtr(const T1& v1, const T2& v2, const T3& v3,
                                  const Allocator& alloc)
{
    allocateData(alloc);
    new(static_cast<Data<Obj_t>*>(dataPtr)) Data<Obj_t>(v1, v2, v3, alloc);
}

template<typename Obj_t, typename Allocator>
template<typename T1, typename T2, typename T3, typename T4>
RefPtr<Obj_t, Allocator>::RefPtr(const T1& v1, const T2& v2, const T3& v3,
                                  const T4& v4, const Allocator& alloc)
{
    allocateData(alloc);
    new(static_cast<Data<Obj_t>*>(dataPtr)) Data<Obj_t>(v1, v2, v3, v4, alloc);
}

template<typename Obj_t, typename Allocator>
template<typename T1, typename T2, typename T3, typename T4, typename T5>
RefPtr<Obj_t, Allocator>::RefPtr(const T1& v1, const T2& v2, const T3& v3,
                                  const T4& v4, const T5& v5,
                                  const Allocator& alloc)
{
    allocateData(alloc);
    new(static_cast<Data<Obj_t>*>(dataPtr)) Data<Obj_t>(v1, v2, v3, v4, v5,
                                                        alloc);
}

template<typename Obj_t, typename Allocator>
template<typename T1, typename T2, typename T3, typename T4, typename T5,
         typename T6>
RefPtr<Obj_t, Allocator>::RefPtr(const T1& v1, const T2& v2, const T3& v3,
                                  const T4& v4, const T5& v5, const T6& v6,
                                  const Allocator& alloc)
{
    allocateData(alloc);
    new(static_cast<Data<Obj_t>*>(dataPtr)) Data<Obj_t>(v1, v2, v3, v4, v5, v6,
                                                        alloc);
}

template<typename Obj_t, typename Allocator>
RefPtr<Obj_t, Allocator>::~RefPtr()
{
    decRefCount();
}

template<typename Obj_t, typename Allocator>
RefPtr<Obj_t, Allocator>::RefPtr(const RefPtr& rhs):
    dataPtr(rhs.dataPtr)
{
    if(dataPtr) ++(dataPtr->refCount);
}

template<typename Obj_t, typename Allocator>
RefPtr<Obj_t, Allocator>&
RefPtr<Obj_t, Allocator>::operator=(const RefPtr& rhs)
{
    if(dataPtr != rhs.dataPtr)
    {
        decRefCount();
        dataPtr = rhs.dataPtr;
        if(dataPtr) ++(dataPtr->refCount);
    }
    return *this;
}

template<typename Obj_t, typename Allocator>
Obj_t* RefPtr<Obj_t, Allocator>::operator->()
{
    return &(static_cast<Data<Obj_t>*>(dataPtr)->data);
}

template<typename Obj_t, typename Allocator>
const Obj_t* RefPtr<Obj_t, Allocator>::operator->() const
{
    return &(static_cast<Data<Obj_t>*>(dataPtr)->data);
}

template<typename Obj_t, typename Allocator>
Obj_t& RefPtr<Obj_t, Allocator>::operator*()
{
    return static_cast<Data<Obj_t>*>(dataPtr)->data;
}

template<typename Obj_t, typename Allocator>
const Obj_t& RefPtr<Obj_t, Allocator>::operator*() const
{
    return static_cast<Data<Obj_t>*>(dataPtr)->data;
}

template<typename Obj_t, typename Allocator>
bool RefPtr<Obj_t, Allocator>::operator==(const RefPtr& rhs) const
{
    return dataPtr == rhs.dataPtr;
}

template<typename Obj_t, typename Allocator>
bool RefPtr<Obj_t, Allocator>::operator!=(const RefPtr& rhs) const
{
    return dataPtr != rhs.dataPtr;
}

template<typename Obj_t, typename Allocator>
Obj_t* RefPtr<Obj_t, Allocator>::get() const
{
    return &(static_cast<Data<Obj_t>*>(dataPtr)->data);
}

template<typename Obj_t, typename Allocator>
RefPtr<Obj_t, Allocator>::operator bool() const
{
    return dataPtr != 0;
}

template<typename Obj_t, typename Allocator>
bool RefPtr<Obj_t, Allocator>::isShared() const
{
    return dataPtr && dataPtr->refCount > 1;
}

#endif
